class
	CODE_TRANSFORMER

inherit
	BINDINGS_STATUS

create
	make

feature {NONE}

	make (a_project_settings: PROJECT_SETTINGS)
		require
			a_project_settings /= Void
			a_project_settings.cil_file /= Void
			a_project_settings.output_dir /= Void
		do
			project_settings := a_project_settings
			create steps.make_from_array (transformer_steps)
			current_step_index := 0
			progress_listener := create {TWO_WAY_LIST [PROCEDURE [ANY, TUPLE [INTEGER]]]}.make
			create tree.make
		ensure
			project_settings = a_project_settings
			steps /= Void and current_step_index = 0
		end

feature {ANY}

	project_settings: PROJECT_SETTINGS

	steps: ARRAY [TRANSFORMATION_STEP]

	current_step_index: INTEGER

	progress_listener: LIST [PROCEDURE [ANY, TUPLE [INTEGER]]]

	go (target_step_index: INTEGER): BOOLEAN
		require
			target_step_index >= 1 and target_step_index <= steps.count
		local
			done, ok: BOOLEAN
		do
			ok := True
			from
				if target_step_index <= current_step_index then
					current_step_index := 1
				else -- target_step_index > current_step_index
					current_step_index := current_step_index + 1
				end
			until
				done
			loop
				if current_step_index = 1 then
					Bindings_active.set_item (False) -- deactivate binding checking
				end
				notify_progress_listener
				ok := steps.item (current_step_index).process (Current)
				if not ok then
					done := True
					current_step_index := 0
					notify_progress_listener
				elseif current_step_index = target_step_index then
					done := True
				else
					current_step_index := current_step_index + 1
				end
			end
			Result := ok
		end

	emit_tree (ev_tree: EV_TREE)
		require
			current_step_index > 0
		do
			steps.item (current_step_index).emit_tree (ev_tree)
		end

	emit_source (text: EV_TEXT)
		require
			current_step_index > 0
		do
			steps.item (current_step_index).emit_source (text)
		end

feature {ANY}

	tree: AST_TRANSLATION_UNIT

	oo_cluster: OO_CLUSTER assign set_oo_cluster

	set_oo_cluster (a_oo_cluster: OO_CLUSTER)
		do
			oo_cluster := a_oo_cluster
		end

feature {NONE}

	notify_progress_listener
		do
			across progress_listener as c loop
				c.item.call ([current_step_index])
			end
		end

	transformer_steps: ARRAY [TRANSFORMATION_STEP]
		local
			list: ARRAYED_LIST [TRANSFORMATION_STEP]
			config: CODE_TRANSFORMER_CONFIG
		do
			create list.make (0)
			create config

			-- create tree:
			list.extend (create {SOURCE_LOADER})
			-- basic code cleanup:
			list.extend (create {TYPEDEF_REMOVER})
			list.extend (create {ENUM_TYPE_TRANSFORMER})
			-- add bindings:
			list.extend (create {FORWARD_DECLARATIONS_REMOVER})
			list.extend (create {BINDING_ADDER})
			-- do transformations:
			list.extend (create {INCREMENT_DECREMENT_TRANSFORMER})
			list.extend (create {SIZEOF_TRANSFORMER})
			list.extend (create {FLUSH_TRANSFORMER})

			if config.run_cil_transformers then
				list.extend (create {CIL_CHAR_TRANSFORMER})
				list.extend (create {CIL_POINTER_ACCESS_TRANSFORMER})
			end

			list.extend (create {ASSIGNMENT_TRANSFORMER})
			list.extend (create {POINTER_ACCESS_INDEX_TRANSFORMER})
			list.extend (create {ARITHMETIC_OPERATION_TRANSFORMER})
			list.extend (create {CAST_REMOVER})
			if config.run_string_transformer then
				list.extend (create {STRING_TRANSFORMER})
			end

			list.extend (create {POINTER_TRANSFORMER})
			list.extend (create {CAST_TRANSFORMER})

			if config.run_printf_transformer then
				list.extend (create {PRINTF_TRANSFORMER})
			end

			list.extend (create {STD_STREAM_TRANSFORMER})
			list.extend (create {LITERAL_TRANSFORMER})
			list.extend (create {BOOLEAN_TRANSFORMER})
			list.extend (create {PARAMETER_ASSIGNMENT_TRANSFORMER})
			list.extend (create {ASSEMBLER_TRANSFORMER})
			list.extend (create {VARIABLE_ADDRESS_TRANSFORMER})
			list.extend (create {ADDRESS_OPERATOR_TRANSFORMER})
			list.extend (create {IDENTIFIER_TRANSFORMER})
--			list.extend (create {CIL_CODE_DUPLICATION_REMOVER})

			if config.run_bcr_transformer then
				list.extend (create {BREAK_CONTINUE_RETURN_TRANSFORMER})
			end

			list.extend (create {LONGJMP_TRANSFORMER})
			list.extend (create {SWITCH_TRANSFORMER})
			list.extend (create {JUMP_STATEMENT_TRANSFORMER})
			list.extend (create {VARARG_TRANSFORMER})
			list.extend (create {ALLOCA_TRANSFORMER})
			list.extend (create {BUILTIN_TRANSFORMER})
			list.extend (create {EXPRESSION_AS_INSTRUCTION_TRANSFORMER})
			list.extend (create {CIL_TEMP_VARIABLE_REMOVER})
			list.extend (create {EXPANDED_TRANSFORMER})
			list.extend (create {STRUCT_TRANSFORMER})

			-- O-O reengineering
			list.extend (create {OO_TRANSFORMER})
			list.extend (create {OO_INHERITANCE_TRANSFORMER})

			-- emit Eiffel source code:
			list.extend (create {EIFFEL_FILE_WRITER})

			Result := list.to_array
		end

invariant
	tree /= Void

end
