class
	AST_STRUCT_DECLARATION

inherit
	AST_DECLARATION
		redefine
			make_copy
		end
	ATTRIBUTE_SPECIFIER_INFO
		redefine
			make_copy
		end

create
	make_from_antlr,
	make_copy

feature {NONE}

	make_from_antlr (node: ANTLR_AST_NODE; a_parent: AST_NODE; a_source_info: SOURCE_INFO)
		require
			node /= Void and then (node.type = C_kw_struct or node.type = C_kw_union)
			a_source_info /= Void
		do
			parent := a_parent

			is_union := node.type = C_kw_union
			fields := create {TWO_WAY_LIST [AST_STRUCT_FIELD]}.make
			is_forward_declaration := True
			across node.children as c loop
				inspect
					c.item.type
				when Ast_variable_declaration_or_type then
					fields.extend (create {AST_VARIABLE_DECLARATION}.make_from_antlr (c.item, Current, Void))
				when Ast_bitfield then
					fields.extend (create {AST_BITFIELD}.make_from_antlr (c.item, Current))
				when C_op_open_curly_brace then
					is_forward_declaration := False
				when C_kw_attribute then
					create gcc_attribute_specifier.make_from_antlr (c.item, Current)
				else
					identifier := c.item.token_text
					orig_c_identifier := identifier.twin
				end
			end
			source_info := a_source_info
		end

feature {ANY}

	is_union: BOOLEAN

	fields: LIST [AST_STRUCT_FIELD]

	is_forward_declaration: BOOLEAN

	orig_c_identifier: STRING

feature {ANY}

	typedef_name: STRING assign set_typedef_name

	set_typedef_name (a_typedef_name: STRING)
		do
			typedef_name := a_typedef_name
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_struct_declaration (Current)
			if visit_children then
				visit_list (fields, v)
				visit_node (gcc_attribute_specifier, v)
			end
			v.end_visit_struct_declaration (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (a_other: like Current)
		do
			Precursor {AST_DECLARATION} (a_other)
			Precursor {ATTRIBUTE_SPECIFIER_INFO} (a_other)
			orig_c_identifier := a_other.orig_c_identifier -- won't be changed
			is_union := a_other.is_union
			fields := create {TWO_WAY_LIST [AST_STRUCT_FIELD]}.make
			across a_other.fields as c loop
				fields.extend (c.item.ast_twin)
				fields.last.parent := Current
			end
			is_forward_declaration := a_other.is_forward_declaration
			typedef_name := a_other.typedef_name
		end

invariant
	fields /= Void
	orig_c_identifier /= Void and then not orig_c_identifier.is_empty

end
