class
	AST_ASSIGNMENT_EXPRESSION

inherit
	AST_EXPRESSION

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_target, a_source: AST_EXPRESSION)
		require
			a_target /= Void and then a_target.parent = Void
			a_source /= Void and then a_target.parent = Void
		do
			type := C_op_assign
			target := a_target
			target.parent := Current
			source := a_source
			source.parent := Current
		end

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_assignment_expression
		do
			parent := init_parent

			target := process_expression (node.children.i_th (1), Current)
			type := node.children.i_th (2).type
			source := process_expression (node.children.i_th(3), Current)
		end

feature {ANY}

	target: AST_EXPRESSION

	type: INTEGER

	source: AST_EXPRESSION

	type_binding: AST_TYPE
		do -- don't use object once here, might change for string type
			Result := target.type_binding
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_assignment_expression (Current)
			if visit_children then
				visit_node (target, v)
				visit_node (source, v)
			end
			v.end_visit_assignment_expression (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			target := other.target.ast_twin
			target.parent := Current
			type := other.type
			source := other.source.ast_twin
			source.parent := Current
		end

invariant
	target /= Void
	type = C_op_assign or type = C_op_mult_assign or type = C_op_div_assign or
			type = C_op_modulo_assign or type = C_op_plus_assign or
			type = C_op_minus_assign or type = C_op_left_shift_assign or
			type = C_op_right_shift_assign or type = C_op_bitand_assign or
			type = C_op_exor_assign or type = C_op_bitor_assign
	source /= Void

end
