class
	AST_DECIMAL_CONSTANT

inherit
	AST_EXPRESSION

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_value: INTEGER; a_type_binding: AST_BASIC_TYPE)
		require
			a_type_binding /= Void
		do
			value := a_value.out
			type_binding := a_type_binding
		ensure
			a_value.out.is_equal (value)
			type_binding = a_type_binding
		end

	make_from_antlr (node: ANTLR_AST_NODE; a_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_decimal_constant
		local
			s: STRING
		do
			parent := a_parent
			s := node.children.first.token_text.as_lower
			if s.ends_with ("ull") then
				value := s.substring (1, s.count - 3)
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_long_long)
			elseif s.ends_with ("ul") then
				value := s.substring (1, s.count - 2)
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_long)
			elseif s.ends_with ("ll") then
				value := s.substring (1, s.count - 2)
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_long_long)
			elseif s.ends_with ("l") then
				value := s.substring (1, s.count - 1)
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_long)
			elseif s.ends_with ("u") then
				value := s.substring (1, s.count - 1)
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_unsigned_int)
			else
				value := s
				type_binding := create {AST_BASIC_TYPE}.make (Ast_bt_int)
			end
		end

feature {ANY}

	value: STRING
		-- don't use "NUMERIC"; assignment in "make_copy" won't work (expanded bug?!)
		-- -> raw number, only digits (see grammar)

	type_binding: AST_BASIC_TYPE

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_decimal_constant (Current)
			if visit_children then
				-- no children
			end
			v.end_visit_decimal_constant (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			value := other.value.twin
			type_binding := other.type_binding
		end

invariant
	value /= Void
	type_binding /= Void and then
	(type_binding.type = Ast_bt_int or type_binding.type = Ast_bt_long or type_binding.type = Ast_bt_long_long or
	type_binding.type = Ast_bt_unsigned_int or type_binding.type = Ast_bt_unsigned_long or type_binding.type = Ast_bt_unsigned_long_long)

end
