class
	AST_GENERIC_EXPRESSION

inherit
	AST_EXPRESSION
	AST_UTILITIES

create
	make,
	make_from_antlr,
	make_copy

feature {NONE}

	make (a_type: INTEGER; a_left, a_right: AST_EXPRESSION)
		require
			a_left /= Void and then a_left.parent = Void
			a_right /= Void and then a_right.parent = Void
		do
			type := a_type
			a_left.parent := Current
			left := a_left
			a_right.parent := Current
			right := a_right
		end

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void
			-- valid node type is checked in the invariant; pretty ugly!
		do
			parent := init_parent
			type := node.type
			node.children.start
			left := process_expression (node.children.item, Current)
			node.children.forth
			right := process_expression (node.children.item, Current)
		ensure
			type = node.type
		end

feature {ANY}

	type: INTEGER

	left: AST_EXPRESSION assign set_left

	right: AST_EXPRESSION assign set_right

	set_left (a_left: AST_EXPRESSION)
		require
			a_left /= Void and then a_left.parent = Void
		do
			left.parent := Void
			a_left.parent := Current
			left := a_left
		end

	set_right (a_right: AST_EXPRESSION)
		require
			a_right /= Void and then a_right.parent = Void
		do
			right.parent := Void
			a_right.parent := Current
			right := a_right
		end

	is_additive_expression: BOOLEAN
		do
			Result := type = C_op_plus or type = C_op_minus
		end

	is_equality_expression: BOOLEAN
		do
			Result := type = C_op_equal or type = C_op_not_equal
		end

	is_relational_expression: BOOLEAN
		do
			Result := type = C_op_smaller or type = C_op_greater or type = C_op_smaller_equal or type = C_op_greater_equal
		end

	is_bit_expression: BOOLEAN
		do
			Result := type = C_op_bitand or type = C_op_bitor or type = C_op_exor
		end

	is_multiplicative_expression: BOOLEAN
		do
			Result := type = C_op_mult or type = C_op_div or type = C_op_modulo
		end

	is_logical_expression: BOOLEAN
		do
			Result := type = C_op_logical_and or type = C_op_logical_or
		end

	is_shift_expression: BOOLEAN
		do
			Result := type = C_op_left_shift or type = C_op_right_shift
		end

	type_binding: AST_TYPE
		local
			left_type, right_type: AST_TYPE
			left_type_basic, right_type_basic: BOOLEAN
			left_type_pointer_or_array, right_type_pointer_or_array: BOOLEAN
		once ("OBJECT")
			if is_equality_expression or is_relational_expression or is_logical_expression then
				Result := create {AST_BASIC_TYPE}.make (Ast_bt_bool)
			else
				left_type := left.type_binding
				right_type := right.type_binding

				left_type_basic := attached {AST_BASIC_TYPE} left_type
				right_type_basic := attached {AST_BASIC_TYPE} right_type
				left_type_pointer_or_array := attached {AST_POINTER_TYPE} left_type or attached {AST_ARRAY_TYPE} left_type
				right_type_pointer_or_array := attached {AST_POINTER_TYPE} right_type or attached {AST_ARRAY_TYPE} right_type

				if left_type_basic and right_type_basic and (is_additive_expression or is_multiplicative_expression) then
					inspect
						({AST_BASIC_TYPE} #? left_type).type
					when Ast_bt_bool then
						Result := create {AST_BASIC_TYPE}.make (Ast_bt_int)
					when
						Ast_bt_int, Ast_bt_long, Ast_bt_long_long,
						Ast_bt_unsigned_int, Ast_bt_unsigned_long, Ast_bt_unsigned_long_long,
						Ast_bt_float, Ast_bt_double, Ast_bt_long_double
					then
						Result := left_type
					end
				elseif left_type_basic and right_type_basic and (is_bit_expression or is_shift_expression) then
					inspect
						({AST_BASIC_TYPE} #? left_type).type
					when Ast_bt_bool then
						Result := create {AST_BASIC_TYPE}.make (Ast_bt_int)
					when
						Ast_bt_int, Ast_bt_long, Ast_bt_long_long,
						Ast_bt_unsigned_int, Ast_bt_unsigned_long, Ast_bt_unsigned_long_long
					then
						Result := left_type
					end
				elseif left_type_pointer_or_array and right_type_basic and is_additive_expression then
					if attached {AST_ARRAY_TYPE} left_type as tmp then
						Result := create {AST_POINTER_TYPE}.make (tmp.type.ast_twin)
					else
						Result := left_type
					end
				elseif left_type_pointer_or_array and right_type_pointer_or_array and is_additive_expression then
					Result := create {AST_BASIC_TYPE}.make (Ast_bt_int)
				else
					Io.error.put_string ("Unsupported AST_GENERIC_EXPRESSION: " + c_code + "%N")
				end
			end
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_generic_expression (Current)
			if visit_children then
				visit_node (left, v)
				visit_node (right, v)
			end
			v.end_visit_generic_expression (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			type := other.type
			left := other.left.ast_twin
			left.parent := Current
			right := other.right.ast_twin
			right.parent := Current
		end

invariant
	is_additive_expression or is_equality_expression or is_relational_expression or
			is_bit_expression or is_multiplicative_expression or is_logical_expression or is_shift_expression
	left /= Void
	right /= Void

end
