class
	AST_POSTFIX_EXPRESSION

inherit
	AST_EXPRESSION
	AST_UTILITIES

create
	make_point,
	make_arrow,
	make_array,
	make_increment,
	make_decrement,
	make_from_antlr,
	make_copy

feature {NONE}

	make_point (a_main_expression: AST_EXPRESSION; a_identifier: STRING; a_type_binding: AST_TYPE; a_declaration_binding: AST_STRUCT_FIELD)
		require
			a_main_expression /= Void
			a_identifier /= Void
			Bindings_active.item implies (a_type_binding /= Void and a_declaration_binding /= Void)
		do
			main_expression := a_main_expression
			main_expression.parent := Current
			type := C_op_point
			identifier := a_identifier
			type_binding := a_type_binding
			declaration_binding := a_declaration_binding
		end

	make_arrow (a_main_expression: AST_EXPRESSION; a_identifier: STRING; a_type_binding: AST_TYPE; a_declaration_binding: AST_STRUCT_FIELD)
		require
			a_main_expression /= Void
			a_identifier /= Void
			Bindings_active.item implies (a_type_binding /= Void and a_declaration_binding /= Void)
		do
			main_expression := a_main_expression
			main_expression.parent := Current
			type := C_op_arrow
			identifier := a_identifier
			type_binding := a_type_binding
			declaration_binding := a_declaration_binding
		end

	make_array (a_main_expression: AST_EXPRESSION; a_array_expression: AST_EXPRESSION)
		require
			a_main_expression /= Void
			a_array_expression /= Void
		do
			main_expression := a_main_expression
			main_expression.parent := Current
			type := C_op_open_square_bracket
			array_expression := a_array_expression
			array_expression.parent := Current
			if bindings_active.item then
				type_binding := deref (main_expression.type_binding)
			end
		end

	make_increment (a_main_expression: AST_EXPRESSION)
		require
			a_main_expression /= Void and then a_main_expression.parent = Void
		do
			main_expression := a_main_expression
			main_expression.parent := Current
			type := C_op_increment
			if bindings_active.item then
				type_binding := main_expression.type_binding
			end
		end

	make_decrement (a_main_expression: AST_EXPRESSION)
		require
			a_main_expression /= Void and then a_main_expression.parent = Void
		do
			main_expression := a_main_expression
			main_expression.parent := Current
			type := C_op_decrement
			if bindings_active.item then
				type_binding := main_expression.type_binding
			end
		end

	make_from_antlr (node: ANTLR_AST_NODE; a_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_postfix_expression
		local
			cur_main_expression: AST_EXPRESSION
		do
			from
				node.children.start
				cur_main_expression := process_expression (node.children.item, Void)
				node.children.forth
			until
				node.children.after
			loop
				inspect
					node.children.item.type
				when C_op_open_square_bracket then
					node.children.forth
					if node.children.index + 1 = node.children.count then
						make_array (cur_main_expression, process_expression (node.children.item, Current))
					else
						cur_main_expression := create {AST_POSTFIX_EXPRESSION}.make_array (cur_main_expression, process_expression (node.children.item, Current))
					end
					node.children.forth
				when C_op_point then
					node.children.forth
					if node.children.islast then
						make_point (cur_main_expression, node.children.item.token_text, Void, Void)
					else
						cur_main_expression := create {AST_POSTFIX_EXPRESSION}.make_point (cur_main_expression, node.children.item.token_text, Void, Void)
					end
				when C_op_arrow then
					node.children.forth
					if node.children.islast then
						make_arrow (cur_main_expression, node.children.item.token_text, Void, Void)
					else
						cur_main_expression := create {AST_POSTFIX_EXPRESSION}.make_arrow (cur_main_expression, node.children.item.token_text, Void, Void)
					end
				when C_op_increment then
					make_increment (cur_main_expression)
				when C_op_decrement then
					make_decrement (cur_main_expression)
				end
				node.children.forth
			end
			parent := a_parent
		end

feature {ANY}

	main_expression: AST_EXPRESSION
			-- a.b.c -> a.b

	base_expression: AST_EXPRESSION
			-- a.b.c -> a
		do
			from
				Result := main_expression.non_parentheses_expression
			until
				not attached {AST_POSTFIX_EXPRESSION} Result as tmp
			loop
				Result := tmp.main_expression.non_parentheses_expression
			end
		end

	type: INTEGER
	identifier: STRING
	array_expression: AST_EXPRESSION

	type_binding: AST_TYPE assign set_type_binding

	set_type_binding (a_type_binding: AST_TYPE)
		require
			a_type_binding /= Void
		do
			type_binding := a_type_binding
		end

	declaration_binding: AST_STRUCT_FIELD assign set_declaration_binding

	set_declaration_binding (a_declaration: AST_STRUCT_FIELD)
		require
			a_declaration /= Void
		do
			declaration_binding := a_declaration
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_postfix_expression (Current)
			if visit_children then
				visit_node (main_expression, v)
				visit_node (array_expression, v)
			end
			v.end_visit_postfix_expression (Current)
			v.post_visit (Current)
		end

	ast_twin: like Current
		do
			create Result.make_copy (Current)
		end

	make_copy (other: like Current)
		do
			main_expression := other.main_expression.ast_twin
			main_expression.parent := Current
			type := other.type
			if other.identifier /= Void then
				identifier := other.identifier.twin
			end
			if other.array_expression /= Void then
				array_expression := other.array_expression.ast_twin
				array_expression.parent := Current
			end
			type_binding := other.type_binding
			declaration_binding := other.declaration_binding
		end

invariant
	main_expression /= Void
	type = C_op_open_square_bracket or type = C_op_point or type = C_op_arrow or type = C_op_increment or type = C_op_decrement;
	(type = C_op_open_square_bracket) = (array_expression /= Void)
	(type = C_op_point or type = C_op_arrow) = (identifier /= Void)
	((type = C_op_point or type = C_op_arrow) and Bindings_active.item and not is_in_orig_c_expression) implies declaration_binding /= Void

end
