class
	AST_IF_ELSE_STATEMENT

inherit
	AST_IF_STATEMENT
		redefine
			accept,
			make_copy
		end

create
	make_if,
	make_if_short,
	make_if_else,
	make_if_else_short,
	make_from_antlr,
	make_copy

feature {NONE}

	make_if_else_short (a_condition: AST_EXPRESSION; a_then, a_else: AST_STATEMENT)
		require
			a_condition /= Void and then a_condition.parent = Void
			a_then /= Void and then a_then.parent = Void
			a_else /= Void and then a_else.parent = Void
		do
			make_if_short (a_condition, a_then)
			create else_statements.make_with_statement (a_else)
			else_statements.parent := Current
		end

	make_if_else (a_condition: AST_EXPRESSION; a_then, a_else: AST_COMPOUND_STATEMENT)
		require
			a_condition /= Void and then a_condition.parent = Void
			a_then /= Void and then a_then.parent = Void
			a_else /= Void and then a_else.parent = Void
		do
			make_if (a_condition, a_then)
			else_statements := a_else
			else_statements.parent := Current
		end

	make_from_antlr (node: ANTLR_AST_NODE; init_parent: AST_NODE)
		require
			node /= Void and then node.type = Ast_if_statement
		local
			statement: AST_STATEMENT
		do
			parent := init_parent

			node.children.start
			condition := process_expression (node.children.item, Current)
			node.children.forth

			statement := process_statement (node.children.item, Void)
			if attached {AST_COMPOUND_STATEMENT} statement as tmp then
				then_statements := tmp
			else
				create then_statements.make_with_statement (statement)
			end
			then_statements.parent := Current
			node.children.forth

			if not node.children.after then
				statement := process_statement (node.children.item, Void)
				if attached {AST_COMPOUND_STATEMENT} statement as tmp then
					else_statements := tmp
				else
					create else_statements.make_with_statement (statement)
				end
				else_statements.parent := Current
			end
		end

feature {ANY}

	else_statements: AST_COMPOUND_STATEMENT assign set_else_statements

	set_else_statements (a_else_statements: AST_COMPOUND_STATEMENT)
		require
			a_else_statements = Void or else a_else_statements.parent = Void
		do
			if a_else_statements /= Void then
				a_else_statements.parent := Current
			end
			else_statements := a_else_statements
		end

feature {ANY} -- AST_NODE

	accept (v: AST_VISITOR)
		local
			visit_children: BOOLEAN
		do
			v.pre_visit (Current)
			visit_children := v.visit_if_statement (Current)
			if visit_children then
				visit_node (condition, v)
				visit_node (then_statements, v)
				visit_node (else_statements, v)
			end
			v.end_visit_if_statement (Current)
			v.post_visit (Current)
		end

	make_copy (other: like Current)
		do
			Precursor (other)
			if other.else_statements /= Void then
				else_statements := other.else_statements.ast_twin
				else_statements.parent := Current
			end
		end

end
