class
	OO_CLUSTER

inherit
	OO_DIRECTORY
		rename
			classes as classes_from_directory,
			get_class as get_class_from_directory,
			has_class as has_class_from_directory,
			add_class as add_class_from_directory
		end

create
	make

feature {ANY}

	classes: LIST [OO_CLASS]
		do
			Result := create {TWO_WAY_LIST [OO_CLASS]}.make
			collect_classes (Result, Current)
		end

	get_class (a_name: STRING_8): OO_CLASS
		require
			a_name /= Void and then not a_name.is_empty
		do
			Result := find_class_in_directory (Current, a_name)
		end

	has_class (a_class: OO_CLASS): BOOLEAN
		require
			a_class /= Void
		do
			Result := get_class (a_class.name) /= Void
		end

	add_class (a_directory: STRING_8; a_class: OO_CLASS)
		require
			a_directory /= Void implies not a_directory.is_empty
			a_class /= Void
			not has_class (a_class)
		do
			add_class_to_directory (Current, a_directory, a_class)
		end

feature {NONE}

	collect_classes (a_class_list: LIST [OO_CLASS]; a_directory: OO_DIRECTORY)
		require
			a_class_list /= Void
			a_directory /= Void
		do
			across a_directory.files as c loop
				if attached {OO_CLASS} c.item as tmp then
					a_class_list.extend (tmp)
				elseif attached {OO_DIRECTORY} c.item as tmp then
					collect_classes (a_class_list, tmp)
				end
			end
		end

	find_class_in_directory (a_directory: OO_DIRECTORY; a_name: STRING_8): OO_CLASS
		require
			a_directory /= Void
			a_name /= Void and then not a_name.is_empty
		do
			across
				a_directory.files as c
			until
				Result /= Void
			loop
				if attached {OO_CLASS} c.item as tmp then
					if tmp.name.is_equal (a_name) then
						Result := tmp
					end
				elseif attached {OO_DIRECTORY} c.item as tmp then
					Result := find_class_in_directory (tmp, a_name)
				end
			end
		end

	add_class_to_directory (a_directory: OO_DIRECTORY; a_sub_dir_name: STRING_8; a_class: OO_CLASS)
		require
			a_directory /= Void
			a_sub_dir_name /= Void implies not a_sub_dir_name.is_empty
			a_class /= Void
		local
			l_index: INTEGER_32
			l_sub_dir_next_name, l_sub_dir_rest_name: STRING_8
			l_sub_dir_next: OO_DIRECTORY
		do
			if a_sub_dir_name = Void then
				a_directory.add_class (a_class)
			else
				l_index := a_sub_dir_name.index_of ('/', 1)
				if l_index = 0 then
					l_sub_dir_next_name := a_sub_dir_name
				else
					l_sub_dir_next_name := a_sub_dir_name.substring (1, l_index - 1)
					l_sub_dir_rest_name := a_sub_dir_name.substring (l_index + 1, a_sub_dir_name.count)
				end

				l_sub_dir_next := a_directory.get_directory (l_sub_dir_next_name)
				if l_sub_dir_next = Void then
					l_sub_dir_next := a_directory.create_directory (l_sub_dir_next_name)
				end

				add_class_to_directory (l_sub_dir_next, l_sub_dir_rest_name, a_class)
			end
		end

end
