class
	CAST_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_cast_expression
		end
	NAME_UTILITIES

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Transforming Casts"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_cast_expression (node: AST_CAST_EXPRESSION)
		local
			type_type, expression_type: AST_TYPE
			e: AST_EXPRESSION
		do
			type_type := node.type
			expression_type := node.expression.type_binding

			-- adapt casting basic types: `(unsigned short)8' -> `(8).to_natural_16'
			if attached {AST_BASIC_TYPE} type_type as bt_dest and attached {AST_BASIC_TYPE} expression_type as bt_src then
				replace_node (node.parentheses_parent, create_basic_cast (node.expression, bt_src, bt_dest))

			-- adapt pointer casts: `(bar *)foo' -> `create {CE_POINTER [BAR]}.make_cast (foo)'
			elseif
				(attached {AST_POINTER_TYPE} type_type or attached {AST_ARRAY_TYPE} type_type) and
				(attached {AST_POINTER_TYPE} expression_type or attached {AST_ARRAY_TYPE} expression_type)
			then
				replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (
						<<"create {", type_type.ast_twin, "}.make_cast (", node.expression.ast_twin, ")" >>,
						type_type))

			-- adapt pointer to basic cast: `(unsigned int)p' -> `(unsigned int)(p).c_pointer.to_integer_32'
			elseif
				attached {AST_BASIC_TYPE} type_type and
				(attached {AST_POINTER_TYPE} expression_type or attached {AST_ARRAY_TYPE} expression_type)
			then
				replace_node (node.expression, create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<<create {AST_PARENTHESIZED_EXPRESSION}.make (node.expression.ast_twin), ".c_pointer.to_integer_32" >>,
							create {AST_BASIC_TYPE}.make (Ast_bt_int)))
				end_visit_cast_expression (node)

			-- adapt "static" pointers:
			-- - `(bar *)0'   -> `create {CE_POINTER [BAR]}.make_shared (default_pointer)'
			-- - `(bar *)123' -> `create {CE_POINTER [BAR]}.make_shared (default_pointer + 123)'
			elseif
				attached {AST_BASIC_TYPE} node.expression.type_binding and then
				(attached {AST_POINTER_TYPE} type_type or attached {AST_ARRAY_TYPE} type_type)
			then
				if attached {AST_DECIMAL_CONSTANT} node.expression as tmp and then tmp.value.is_equal ("0") then
					replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<<"create {", type_type.ast_twin, "}.make_shared (default_pointer)">>,
							type_type))
				else
					e := node.expression.ast_twin
					if
						attached {AST_BASIC_TYPE} e.type_binding as bt and then
						not (bt.type = Ast_bt_int or bt.type = Ast_bt_long)
					then
						e := create {AST_CAST_EXPRESSION}.make (create {AST_BASIC_TYPE}.make (Ast_bt_int), e)
					end
					replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<<"create {", type_type.ast_twin, "}.make_shared (default_pointer + ", e, ")">>,
							type_type))
					if attached {AST_CAST_EXPRESSION} e as ce then
						end_visit_cast_expression (ce)
					end
				end
			end
		end

feature {NONE}

	create_basic_cast (expression: AST_EXPRESSION; src_type, dest_type: AST_BASIC_TYPE): AST_EXPRESSION
		do
			if
				attached {HASH_TABLE [STRING_8, INTEGER_32]} basic_cast_table[src_type.type] as ht and then
				attached ht[dest_type.type] as s
			then
				if s[1] = 'r' then
					Result := create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<<s + " (", expression.ast_twin, ")">>,
							dest_type
						)
				else
					Result := create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<<create {AST_PARENTHESIZED_EXPRESSION}.make (expression.ast_twin), "." + s>>,
							dest_type
						)
				end
			else
				log ("Basic cast missing: " + src_type.eiffel_code + "->" + dest_type.eiffel_code)
				Result := create {AST_EIFFEL_EXPRESSION}.make_mixed (
						<<create {AST_PARENTHESIZED_EXPRESSION}.make (expression.ast_twin), ".missing_cast">>,
						dest_type
					)
			end
		end

	basic_cast_table: HASH_TABLE [HASH_TABLE [STRING_8, INTEGER_32], INTEGER_32]
		local
			tmp: HASH_TABLE [STRING_8, INTEGER_32]
		once("PROCESS")
			create Result.make (16)

			-- INTEGER_8
			create tmp.make (16)
			Result[Ast_bt_char] := tmp
			tmp[Ast_bt_char]      := "to_integer_8"
			tmp[Ast_bt_short]     := "to_integer_16"
			tmp[Ast_bt_int]       := "to_integer_32"
			tmp[Ast_bt_long]      := "to_integer_32"
			tmp[Ast_bt_long_long] := "to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "to_natural_64"
			tmp[Ast_bt_float]  := "to_real"
			tmp[Ast_bt_double] := "to_double"
			tmp[Ast_bt_long_double] := "real_96_from_integer_8"
			tmp[Ast_bt_bool] := "to_boolean"

			-- INTEGER_16
			create tmp.make (16)
			Result[Ast_bt_short] := tmp
			tmp[Ast_bt_char]      := "to_integer_8"
			tmp[Ast_bt_short]     := "to_integer_16"
			tmp[Ast_bt_int]       := "to_integer_32"
			tmp[Ast_bt_long]      := "to_integer_32"
			tmp[Ast_bt_long_long] := "to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "to_natural_64"
			tmp[Ast_bt_float]  := "to_real"
			tmp[Ast_bt_double] := "to_double"
			tmp[Ast_bt_long_double] := "real_96_from_integer_16"
			tmp[Ast_bt_bool] := "to_boolean"

			-- INTEGER_32
			create tmp.make (16)
			Result[Ast_bt_int] := tmp
			Result[Ast_bt_long] := tmp
			tmp[Ast_bt_char]      := "to_integer_8"
			tmp[Ast_bt_short]     := "to_integer_16"
			tmp[Ast_bt_int]       := "to_integer_32"
			tmp[Ast_bt_long]      := "to_integer_32"
			tmp[Ast_bt_long_long] := "to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "to_natural_64"
			tmp[Ast_bt_float]  := "to_real"
			tmp[Ast_bt_double] := "to_double"
			tmp[Ast_bt_long_double] := "real_96_from_integer_32"
			tmp[Ast_bt_bool] := "to_boolean"

			-- INTEGER_64
			create tmp.make (16)
			Result[Ast_bt_long_long] := tmp
			tmp[Ast_bt_char]      := "to_integer_8"
			tmp[Ast_bt_short]     := "to_integer_16"
			tmp[Ast_bt_int]       := "to_integer_32"
			tmp[Ast_bt_long]      := "to_integer_32"
			tmp[Ast_bt_long_long] := "to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "to_natural_64"
			tmp[Ast_bt_float]  := "to_real"
			tmp[Ast_bt_double] := "to_double"
			tmp[Ast_bt_long_double] := "real_96_from_integer_64"
			tmp[Ast_bt_bool] := "to_boolean"

			-- NATURAL_8
			create tmp.make (16)
			Result[Ast_bt_unsigned_char] := tmp
			tmp[Ast_bt_char]      := "to_integer_8"
			tmp[Ast_bt_short]     := "to_integer_16"
			tmp[Ast_bt_int]       := "to_integer_32"
			tmp[Ast_bt_long]      := "to_integer_32"
			tmp[Ast_bt_long_long] := "to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "to_natural_64"
			tmp[Ast_bt_float]  := "to_real_32"
			tmp[Ast_bt_double] := "to_real_64"
			tmp[Ast_bt_long_double] := "real_96_from_natural_8"
			tmp[Ast_bt_bool] := "to_boolean"

			-- NATURAL_16
			create tmp.make (16)
			Result[Ast_bt_unsigned_short] := tmp
			tmp[Ast_bt_char]      := "to_integer_8"
			tmp[Ast_bt_short]     := "to_integer_16"
			tmp[Ast_bt_int]       := "to_integer_32"
			tmp[Ast_bt_long]      := "to_integer_32"
			tmp[Ast_bt_long_long] := "to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "to_natural_64"
			tmp[Ast_bt_float]  := "to_real_32"
			tmp[Ast_bt_double] := "to_real_64"
			tmp[Ast_bt_long_double] := "real_96_from_natural_16"
			tmp[Ast_bt_bool] := "to_boolean"

			-- NATURAL_32
			create tmp.make (16)
			Result[Ast_bt_unsigned_int] := tmp
			Result[Ast_bt_unsigned_long] := tmp
			tmp[Ast_bt_char]      := "to_integer_8"
			tmp[Ast_bt_short]     := "to_integer_16"
			tmp[Ast_bt_int]       := "to_integer_32"
			tmp[Ast_bt_long]      := "to_integer_32"
			tmp[Ast_bt_long_long] := "to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "to_natural_64"
			tmp[Ast_bt_float]  := "to_real_32"
			tmp[Ast_bt_double] := "to_real_64"
			tmp[Ast_bt_long_double] := "real_96_from_natural_32"
			tmp[Ast_bt_bool] := "to_boolean"

			-- NATURAL_64
			create tmp.make (16)
			Result[Ast_bt_unsigned_long_long] := tmp
			tmp[Ast_bt_char]      := "to_integer_8"
			tmp[Ast_bt_short]     := "to_integer_16"
			tmp[Ast_bt_int]       := "to_integer_32"
			tmp[Ast_bt_long]      := "to_integer_32"
			tmp[Ast_bt_long_long] := "to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "to_natural_64"
			tmp[Ast_bt_float]  := "to_real_32"
			tmp[Ast_bt_double] := "to_real_64"
			tmp[Ast_bt_long_double] := "real_96_from_natural_64"
			tmp[Ast_bt_bool] := "to_boolean"

			-- REAL_32
			create tmp.make (16)
			Result[Ast_bt_float] := tmp
			tmp[Ast_bt_char]      := "truncated_to_integer.to_integer_8"
			tmp[Ast_bt_short]     := "truncated_to_integer.to_integer_16"
			tmp[Ast_bt_int]       := "truncated_to_integer"
			tmp[Ast_bt_long]      := "truncated_to_integer"
			tmp[Ast_bt_long_long] := "truncated_to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "truncated_to_integer_64.to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "truncated_to_integer_64.to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "truncated_to_integer_64.to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "truncated_to_integer_64.to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "truncated_to_integer_64.to_natural_64"
			tmp[Ast_bt_float]  := "item"
			tmp[Ast_bt_double] := "to_double"
			tmp[Ast_bt_long_double] := "real_96_from_real_32"

			-- REAL_64
			create tmp.make (16)
			Result[Ast_bt_double] := tmp
			tmp[Ast_bt_char]      := "truncated_to_integer.to_integer_8"
			tmp[Ast_bt_short]     := "truncated_to_integer.to_integer_16"
			tmp[Ast_bt_int]       := "truncated_to_integer"
			tmp[Ast_bt_long]      := "truncated_to_integer"
			tmp[Ast_bt_long_long] := "truncated_to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "truncated_to_integer_64.to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "truncated_to_integer_64.to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "truncated_to_integer_64.to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "truncated_to_integer_64.to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "truncated_to_integer_64.to_natural_64"
			tmp[Ast_bt_float]  := "truncated_to_real"
			tmp[Ast_bt_double] := "item"
			tmp[Ast_bt_long_double] := "real_96_from_real_64"

			-- REAL_96
			create tmp.make (16)
			Result[Ast_bt_long_double] := tmp
			tmp[Ast_bt_char]      := "truncated_to_integer.to_integer_8"
			tmp[Ast_bt_short]     := "truncated_to_integer.to_integer_16"
			tmp[Ast_bt_int]       := "truncated_to_integer"
			tmp[Ast_bt_long]      := "truncated_to_integer"
			tmp[Ast_bt_long_long] := "truncated_to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "truncated_to_integer_64.to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "truncated_to_integer_64.to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "truncated_to_integer_64.to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "truncated_to_integer_64.to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "truncated_to_integer_64.to_natural_64"
			tmp[Ast_bt_float]  := "truncated_to_real_32"
			tmp[Ast_bt_double] := "truncated_to_real_64"
			tmp[Ast_bt_long_double] := "item"

			-- BOOLEAN
			create tmp.make (16)
			Result[Ast_bt_bool] := tmp
			tmp[Ast_bt_char]      := "to_integer.to_integer_8"
			tmp[Ast_bt_short]     := "to_integer.to_integer_16"
			tmp[Ast_bt_int]       := "to_integer"
			tmp[Ast_bt_long]      := "to_integer"
			tmp[Ast_bt_long_long] := "to_integer.to_integer_64"
			tmp[Ast_bt_unsigned_char]      := "to_integer.to_natural_8"
			tmp[Ast_bt_unsigned_short]     := "to_integer.to_natural_16"
			tmp[Ast_bt_unsigned_int]       := "to_integer.to_natural_32"
			tmp[Ast_bt_unsigned_long]      := "to_integer.to_natural_32"
			tmp[Ast_bt_unsigned_long_long] := "to_integer.to_natural_64"
			tmp[Ast_bt_float]  := "to_integer.to_real"
			tmp[Ast_bt_double] := "to_integer.to_double"
			tmp[Ast_bt_long_double] := "real_96_from_boolean"
			tmp[Ast_bt_bool] := "item"
		end

end
