class
	CIL_CODE_DUPLICATION_REMOVER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_if_statement
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			from
				duplication_removed := True
			until
				not duplication_removed
			loop
				duplication_removed := False
				transformer.tree.accept (Current)
			end
			Result := True
		end

	description : STRING
		do
			Result := "Reverting CILs code duplication"
		end

feature {NONE}

	duplication_removed: BOOLEAN

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_if_statement (node: AST_IF_STATEMENT)
		local
			tmp_statement: AST_STATEMENT
		do
			-- if(foo) {
			--    abc
			-- } else {          if(foo || bar) {
			--    if(bar) {           abc
			--       abc     ->  } else {
			--    } else {          cde
			--       cde         }
			--    }
			-- }
			if
				node.else_statements /= Void and then node.else_statements.statements.count = 1 and then
				attached {AST_IF_STATEMENT}node.else_statements.statements.first as sub_if and then
				node.then_statements.c_code.is_equal (sub_if.then_statements.c_code)
			then
				replace_node (node.condition, create {AST_GENERIC_EXPRESSION}.make (C_op_logical_or, clone_node (node.condition), clone_node (sub_if.condition)))
				if sub_if.else_statements /= Void then
					node.else_statements := clone_node (sub_if.else_statements)
				else
					node.else_statements := Void
				end
				duplication_removed := True

			-- if(foo) {
			--    abc         if(foo || bar) {
			-- }          ->     abc
			-- if(bar) {      }
			--    abc
			-- }
			elseif
				node.else_statements = Void and then
				attached {AST_IF_STATEMENT}get_adjacent_statement (node, False) as previous_if and then
				previous_if.else_statements = Void and then
				node.then_statements.c_code.is_equal (previous_if.then_statements.c_code)
			then
				replace_node (node.condition, create {AST_GENERIC_EXPRESSION}.make (C_op_logical_or, clone_node (previous_if.condition), clone_node (node.condition)))
				delete_node (previous_if)
				duplication_removed := True

			-- if(foo) {
			--    if(bar) {      if(foo && bar) {
			--      abc      ->     abc
			--    }              }
			-- }
			elseif
				node.else_statements = Void and then
				node.then_statements.statements.count = 1 and then
				attached {AST_IF_STATEMENT}node.then_statements.statements.first as sub_if and then
				sub_if.else_statements = Void
			then
				replace_node (node.condition, create {AST_GENERIC_EXPRESSION}.make (C_op_logical_and, clone_node (node.condition), clone_node (sub_if.condition)))
				-- replace sub_if with abcd
				tmp_statement := node.then_statements.statements.first
				node.then_statements.statements.wipe_out
				tmp_statement.parent := Void
				across sub_if.then_statements.statements as statement_cursor loop
					tmp_statement := clone_node (statement_cursor.item)
					tmp_statement.parent := node.then_statements
					node.then_statements.statements.extend (tmp_statement)
				end
				duplication_removed := True

			-- it's not that easy dude!
--			-- if(foo) { }  ->  if(foo) { }
--			-- if(bar) { }  ->  else if(bar) { }
--			elseif
--				attached {AST_IF_STATEMENT}get_adjacent_statement (node, False) as previous_if and then
--				attached {AST_IF_STATEMENT}get_last_elseif (previous_if) as last_elseif
--			then
--				last_elseif.else_statements := create {AST_COMPOUND_STATEMENT}.make_with_statement (clone_node (node))
--				delete_node (node)
--				duplication_removed := True
			end
		end

--feature {NONE}

--	get_last_elseif (node: AST_IF_STATEMENT): AST_IF_STATEMENT
--		require
--			node /= Void
--		do
--			if node.else_statements = Void then
--				Result := node
--			elseif
--				node.else_statements.statements.count = 1 and then
--				attached {AST_IF_STATEMENT}node.else_statements.statements.first as sub_if
--			then
--				Result := get_last_elseif (sub_if)
--			end
--		ensure
--			Result /= Void implies Result.else_statements = Void
--		end

end
