class
	CIL_POINTER_ACCESS_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_unary_expression,
			end_visit_generic_expression
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Reverting CILs pointer access adaptions"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_unary_expression (node: AST_UNARY_EXPRESSION)
		local
			array_expression: AST_DECIMAL_CONSTANT
		do
			-- `*foo' -> `foo[0]' (only for string types)
			if
				node.type = c_op_mult and then
				attached {AST_VARIABLE_EXPRESSION} node.expression as expr and then
				is_string_type (expr.type_binding)
			then
				create array_expression.make (0, create {AST_BASIC_TYPE}.make (Ast_bt_int))
				replace_node (node, create {AST_POSTFIX_EXPRESSION}.make_array (expr.ast_twin, array_expression))
			end
		end

	end_visit_generic_expression (node: AST_GENERIC_EXPRESSION)
		local
			array_expression: AST_EXPRESSION
		do
			-- `*(foo + bar)' -> `foo[bar]'
			-- `*(foo - bar)' -> `foo[-bar]'
			if
				node.is_additive_expression and then
				(attached {AST_POINTER_TYPE} node.left.type_binding or attached {AST_ARRAY_TYPE} node.right.type_binding) and then
				attached {AST_UNARY_EXPRESSION} node.non_parentheses_parent as deref_expression and then
				deref_expression.type = C_op_mult
			then
				array_expression := node.right.ast_twin
				if node.type = C_op_minus then
					array_expression := create {AST_UNARY_EXPRESSION}.make (C_op_minus, array_expression)
				end
				replace_node (deref_expression, create {AST_POSTFIX_EXPRESSION}.make_array (node.left.ast_twin, array_expression))
			end
		end

end
