class
	CIL_TEMP_VARIABLE_REMOVER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			visit_function_declaration,
			end_visit_function_declaration,
			end_visit_variable_expression
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description : STRING
		do
			Result := "Removing CILs temporary variables"
		end

feature {NONE}

	variable_usages: HASH_TABLE [LIST [AST_VARIABLE_EXPRESSION], STRING]

feature {ANY} -- AST_DEFAULT_VISITOR

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN
		do
			if node.is_regular then
				create variable_usages.make (node.locals.count)
				Result := True
			end
		end

	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION)
		local
			nodes_to_delete: TWO_WAY_LIST [AST_NODE]
		do
			if node.is_regular then
				create nodes_to_delete.make
				across variable_usages as c loop -- loop through `variable_usages' to process in correct order, not `node.locals'
					if
						c.item.count = 2 and then
						attached {AST_ASSIGNMENT_EXPRESSION} c.item.first.parent as assignment_expression and then
						assignment_expression.target = c.item.first and then
						attached {AST_EXPRESSION_STATEMENT} assignment_expression.parent as expression_statement and then
						-- only remove the temp variable if the two statements follow each other (not safe otherwise):
						expression_statement = get_adjacent_statement (get_parent_statement (c.item.last), False)
					then
						replace_node (c.item.last, create {AST_PARENTHESIZED_EXPRESSION}.make (assignment_expression.source.ast_twin))
						nodes_to_delete.extend (expression_statement)
						nodes_to_delete.extend (c.item.first.declaration_binding)
					end
				end
				across nodes_to_delete as c loop
					delete_node (c.item)
				end
				variable_usages := Void
			end
		end

	end_visit_variable_expression (node: AST_VARIABLE_EXPRESSION)
		do
			if
				attached {AST_VARIABLE_DECLARATION} node.declaration_binding as vd and then
				vd.orig_c_identifier.starts_with ("tmp")
			then
				if not variable_usages.has (vd.orig_c_identifier) then
					variable_usages[vd.orig_c_identifier] := create {TWO_WAY_LIST [AST_VARIABLE_EXPRESSION]}.make
				end
				variable_usages[vd.orig_c_identifier].extend (node)
			end
		end

end
