class
	EIFFEL_FILE_WRITER

inherit
	TRANSFORMATION_STEP
		redefine
			emit_source
		end
	NAME_UTILITIES
	ARGUMENTS

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		local
			main_function: AST_FUNCTION_DECLARATION
			assembler_statements: LIST [AST_ASSEMBLER_STATEMENT]
			base_class_name, launcher_class_name, data_class_name, assembler_class_name, initializer_class_name: STRING
			classes: LIST [OO_CLASS]
			has_initializers: BOOLEAN
		do
			across
				transformer.tree.declarations as c
			until
				main_function /= Void
			loop
				if
					attached {AST_FUNCTION_DECLARATION} c.item as tmp and then
					tmp.identifier.is_equal ("main")
				then
					main_function := tmp
				end
			end
			if main_function /= Void then
				log ("Function %"main%" found, compiling as program.")
			else
				log ("Function %"main%" not found, compiling as library.")
			end
			assembler_statements := (create {ASSEMBLER_TRANSFORMER}).assembler_statements

			base_class_name := transformer.oo_cluster.name.as_upper
			launcher_class_name := base_class_name + "_LAUNCHER"
			data_class_name := base_class_name + "_DATA"
			assembler_class_name := base_class_name + "_ASSEMBLER"
			initializer_class_name := base_class_name + "_INITIALIZER"

			-- writing the classes
			classes := transformer.oo_cluster.classes
			write_classes (transformer.project_settings.output_dir, transformer.oo_cluster, data_class_name, classes)

			-- adding project framework classes
			write_ecf_file (launcher_class_name, base_class_name, main_function /= Void)
			if main_function /= Void then
				write_project_launcher_class (main_function, launcher_class_name, data_class_name)
			end
			if not assembler_statements.is_empty then
				write_assembler_class (assembler_statements, assembler_class_name)
			end
			has_initializers := write_initializer_class (initializer_class_name, data_class_name, classes)
			write_project_data_class (data_class_name, classes,
					not assembler_statements.is_empty, assembler_class_name,
					has_initializers, initializer_class_name
				)
			write_header_file (base_class_name)
			if attached resources_directory as rd then
				copy_utilities (rd.name + "/ce_utilities", "ce_utilities")
				if {PLATFORM}.is_unix then
					copy_utilities (rd.name + "/include/linux", "include")
					copy_utilities (rd.name + "/objects/linux", "objects")
				elseif {PLATFORM}.is_windows then
					copy_utilities (rd.name + "/include/windows", "include")
					copy_utilities (rd.name + "/objects/windows", "objects")
				end
				write_ce_settings_class (create {DIRECTORY}.make (transformer.project_settings.output_dir.name + "/ce_utilities"))
			else
				log ("Can't copy utilities, %"resources%" directory not found.")
			end

			Result := True
		end

	description: STRING
		do
			Result := "Writing Eiffel File(s)"
		end

	emit_source (text: EV_TEXT)
		do
			text.set_text ("All source files are in %"" + transformer.project_settings.output_dir.name + "%".")
		end

feature {NONE}

	write_classes (a_real_dir: DIRECTORY; a_oo_dir: OO_DIRECTORY; a_data_class_name: STRING_8; a_classes: LIST [OO_CLASS])
		require
			a_real_dir /= Void
			a_oo_dir /= Void
			a_data_class_name /= Void
			a_classes /= Void
		local
			l_file: PLAIN_TEXT_FILE
			l_struct_writer: STRUCT_WRITER
		do
			if not a_real_dir.exists then
				a_real_dir.create_dir
			end

			across a_oo_dir.files as c loop
				if attached {OO_REGULAR_CLASS} c.item as tmp then
					if tmp.is_system_source then
						write_system_class (tmp, a_real_dir)
					else
						write_project_class (tmp, a_real_dir, a_data_class_name, a_classes)
					end
				elseif attached {OO_DATA_CLASS} c.item as tmp then
					create l_struct_writer.make_oo (tmp)
					l_struct_writer.write

					create l_file.make_open_write (a_real_dir.name + "/" + eiffel_file_name (tmp.name))
					l_file.put_string (l_struct_writer.buffer)
					l_file.flush
					l_file.close
				elseif attached {OO_CONSTANT_CLASS} c.item as tmp then
					create l_file.make_open_write (a_real_dir.name + "/" + eiffel_file_name (tmp.name))
					l_file.put_string (tmp.node.eiffel_code)
					l_file.flush
					l_file.close
				elseif attached {OO_DIRECTORY} c.item as tmp then
					write_classes (create {DIRECTORY}.make (a_real_dir.name + "/" + tmp.name), tmp, a_data_class_name, a_classes)
				end
			end
		end

	write_system_class (a_class: OO_REGULAR_CLASS; a_dir: DIRECTORY)
		require
			a_class /= Void and then a_class.is_system_source
			a_dir /= Void
		local
			l_attributes, l_routines: LIST [OO_FEATURE]
			l_file: PLAIN_TEXT_FILE
		do
			l_attributes := a_class.attributes
			l_routines := a_class.routines
			create l_file.make_open_write (a_dir.name + "/" + eiffel_file_name (a_class.name))

			l_file.put_string ("deferred class%N")
			l_file.put_string ("	" + a_class.name + "%N%N")

			l_file.put_string ("inherit%N")
			l_file.put_string ("	ANY%N")
			l_file.put_string ("		undefine%N")
			l_file.put_string ("			default_create%N")
			l_file.put_string ("		end%N%N")

			if not l_attributes.is_empty then
				l_file.put_string ("feature {ANY} -- attributes%N%N")
				across l_attributes as c loop
					l_file.put_string ("	")
					l_file.put_string (c.item.node.eiffel_code)
					l_file.put_string ("%N%N")
				end
			end
			if not l_routines.is_empty then
				l_file.put_string ("feature {ANY} -- routines%N%N")
				across l_routines as c loop
					l_file.put_string (c.item.node.eiffel_code)
					l_file.put_string ("%N%N")
				end
			end

			l_file.put_string ("end%N")
			l_file.flush
			l_file.close
		end

	write_project_class (
				a_class: OO_REGULAR_CLASS; a_dir: DIRECTORY;
				a_data_class_name: STRING_8; a_classes: LIST [OO_CLASS]
			)
		require
			a_class /= Void and then not a_class.is_system_source
			a_dir /= Void
		local
			l_attributes, l_routines: LIST [OO_FEATURE]
			l_file: PLAIN_TEXT_FILE
			cur: AST_ARRAY_TYPE
		do
			l_attributes := a_class.attributes
			l_routines := a_class.routines
			create l_file.make_open_write (a_dir.name + "/" + eiffel_file_name (a_class.name))

			l_file.put_string ("class%N")
			l_file.put_string ("	" + a_class.name + "%N%N")

			l_file.put_string ("inherit%N")
			l_file.put_string ("	" + a_data_class_name + "%N")
			if not l_attributes.is_empty then
				l_file.put_string ("		redefine%N")
				l_file.put_string ("			default_create%N")
				l_file.put_string ("		end%N")
			end
			across a_classes as c loop
				-- FIXME: only actually required classes should be added
				if
					attached {OO_REGULAR_CLASS} c.item and
					c.item.is_system_source
				then
					l_file.put_string ("	" + c.item.name + "%N")
				end
			end
			l_file.put_string ("%N")

			if not l_attributes.is_empty then
				l_file.put_string ("feature {NONE} -- Initialization%N%N")
				l_file.put_string ("	default_create%N")
				l_file.put_string ("		do%N")
				across l_attributes as c loop
					if
						(
							attached {AST_POINTER_TYPE} c.item.node.type or
							attached {AST_ARRAY_TYPE} c.item.node.type or
							attached {AST_STRUCT_TYPE} c.item.node.type
						) and
						not c.item.node.is_extern
					then
						l_file.put_string ("			create " + c.item.node.identifier + ".make")
						if attached {AST_ARRAY_TYPE} c.item.node.type as array_type then
							l_file.put_string (" (<<")
							from
								cur := array_type
							until
								cur = Void
							loop
								if cur /= array_type then
									l_file.put_string (", ")
								end
								if cur.dimension /= Void then
									l_file.put_string (cur.dimension.eiffel_code)
								else
									l_file.put_string ("TODO") -- not possible?!
								end
								cur ?= cur.type
							end
							l_file.put_string (">>)")
						end
						l_file.put_string ("%N")
					end
				end
				l_file.put_string ("		end%N%N")

				l_file.put_string ("feature {ANY} -- attributes%N%N")
				across l_attributes as c loop
					l_file.put_string ("	")
					l_file.put_string (c.item.node.eiffel_code)
					l_file.put_string ("%N%N")
				end
			end
			if not l_routines.is_empty then
				l_file.put_string ("feature {ANY} -- routines%N%N")
				across l_routines as c loop
					l_file.put_string (c.item.node.eiffel_code)
					l_file.put_string ("%N%N")
				end
			end

			l_file.put_string ("end%N")
			l_file.flush
			l_file.close
		end

	write_assembler_class (assembler_statements: LIST [AST_ASSEMBLER_STATEMENT]; assembler_class_name: STRING)
		require
			assembler_statements /= Void and then not assembler_statements.is_empty
		local
			file: PLAIN_TEXT_FILE
			c_creator: AST_C_CREATOR
			l_type: AST_TYPE
		do
			create file.make_open_write (transformer.project_settings.output_dir.name + "/" + eiffel_file_name (assembler_class_name))
			file.put_string ("class%N")
			file.put_string ("	" + assembler_class_name + "%N%N")

			file.put_string ("feature {NONE}%N%N")

			create c_creator.make
			c_creator.indent := 4
			across assembler_statements as c loop
				file.put_string ("	asm" + c.target_index.out)
				if not c.item.accessed_variables.is_empty then
					file.put_string (" (")
					across c.item.accessed_variables as c2 loop
						if c2.target_index > 1 then
							file.put_string ("; ")
						end
						file.put_string ("op")
						file.put_integer_32 (c2.target_index)
						file.put_string (": ")
						file.put_string (c2.item.type_binding.eiffel_code)
					end
					file.put_string (")%N")
					file.put_string ("		do%N")
					file.put_string ("			c_asm" + c.target_index.out + " (")
					across c.item.accessed_variables as c2 loop
						if c2.target_index > 1 then
							file.put_string (", ")
						end
						file.put_string ("op")
						file.put_integer_32 (c2.target_index)
						file.put_string (".c_pointer")
					end
					file.put_string (")%N")
					file.put_string ("		end%N%N")

					file.put_string ("	c_asm" + c.target_index.out)
					file.put_string (" (")
					across c.item.accessed_variables as c2 loop
						if c2.target_index > 1 then
							file.put_string ("; ")
						end
						file.put_string ("op")
						file.put_integer_32 (c2.target_index)
						file.put_string (": POINTER")
					end
					file.put_string (")")
				end
				file.put_string ("%N")
				file.put_string ("		external%N")
				file.put_string ("			%"C inline use <" + c_header_file (transformer.tree) + ">%"%N")
				file.put_string ("		alias%N")
				file.put_string ("			%"[%N")

				if not c.item.accessed_variables.is_empty then
					across c.item.accessed_variables as c2 loop
						if attached {AST_VARIABLE_DECLARATION} c2.item.declaration_binding as vd then
							l_type := ({AST_ARRAY_TYPE} #? vd.type).type

							-- int x = *(int *)$opX;
							file.put_string ("				")
							if attached {AST_ARRAY_TYPE} l_type as at then
								file.put_string ((create {AST_VARIABLE_DECLARATION}.make (at.as_pointer, vd.orig_c_identifier, Void)).c_code)
							else
								file.put_string ((create {AST_VARIABLE_DECLARATION}.make (l_type.ast_twin, vd.orig_c_identifier, Void)).c_code)
							end

							file.put_string (" = ")

							if attached {AST_ARRAY_TYPE} l_type then
								-- we used `arr.address' what is not really an address; don't dereference it
							else
								file.put_string ("*")
							end
							file.put_string ("(")
							if attached {AST_ARRAY_TYPE} l_type as at then
								file.put_string (at.as_pointer.c_code)
								-- no additional pointer needed; see comment above
							else
								file.put_string (l_type.c_code)
								file.put_string (" *")
							end
							file.put_string (")$op")
							file.put_integer_32 (c2.target_index)
							file.put_string (";%N")
						else
							file.put_string ("?" + c2.item.declaration_binding.generator + "?")
						end
					end
					file.put_string ("%N")
				end

				c.item.accept (c_creator)
				file.put_string ("				")
				file.put_string (c_creator.buffer)
				file.put_string ("%N")
				c_creator.buffer.wipe_out

				if not c.item.accessed_variables.is_empty then
					file.put_string ("%N")
					across c.item.accessed_variables as c2 loop
						if attached {AST_VARIABLE_DECLARATION} c2.item.declaration_binding as vd then
							l_type := ({AST_ARRAY_TYPE} #? vd.type).type

							if attached {AST_ARRAY_TYPE} l_type then
								-- no need to save back; see comment above
							else
								-- *(int *)$x = x;
								file.put_string ("				*(")
								file.put_string (remove_const_qualifiers (l_type).c_code)
								file.put_string (" *)$op")
								file.put_integer_32 (c2.target_index)
								file.put_string (" = ")
								file.put_string (vd.orig_c_identifier)
								file.put_string (";%N")
							end
						else
							file.put_string ("?" + c2.item.declaration_binding.generator + "?")
						end
					end
				end

				file.put_string ("			]%"%N")
				file.put_string ("		end%N%N")
			end

			file.put_string ("end%N")
			file.flush
			file.close
		end

	write_initializer_class (initializer_class_name, data_class_name: STRING; classes: LIST [OO_CLASS]): BOOLEAN
		require
			initializer_class_name /= Void
			data_class_name /= Void
		local
			fake_parent: AST_DECLARATION
			compound_statement: AST_COMPOUND_STATEMENT
			eiffel_creator: AST_EIFFEL_CREATOR
			slice_size, i, j: INTEGER_32
			file: PLAIN_TEXT_FILE
			slices_dir: DIRECTORY
			tmp: TWO_WAY_LIST [AST_STATEMENT]
		do
			-- assignments need to be in the tree, at least for CLASS_UTILITIES.get_class_prefix
			fake_parent := create {AST_VARIABLE_DECLARATION}.make (create {AST_BASIC_TYPE}.make (Ast_bt_int), "_", Void)
			fake_parent.source_info := create {SOURCE_INFO}.make ("#line 1 %"_%"")
			fake_parent.parent := transformer.tree

			compound_statement := create {AST_COMPOUND_STATEMENT}.make
			compound_statement.parent := fake_parent

			across transformer.tree.declarations as c loop
				if attached {AST_VARIABLE_DECLARATION} c.item as vd and then vd.initializer /= Void then
					create_initializer_assignments (
							compound_statement,
							create {AST_VARIABLE_EXPRESSION}.make (vd.identifier, vd),
							vd.initializer
						)
				end
			end

			Result := not compound_statement.statements.is_empty
			if Result then
				create eiffel_creator.make
				eiffel_creator.indent := 2
				slice_size := 500
				i := (compound_statement.statements.count / slice_size).ceiling
				if i = 1 then
					create file.make_open_write (transformer.project_settings.output_dir.name + "/" + eiffel_file_name (initializer_class_name))
					file.put_string ("frozen class%N")
					file.put_string ("	" + initializer_class_name + "%N%N")
					file.put_string ("inherit%N")
					file.put_string ("	" + data_class_name + "%N")
					across classes as c loop
						-- FIXME: only actually required classes should be added
						if
							attached {OO_REGULAR_CLASS} c.item and
							c.item.is_system_source
						then
							file.put_string ("	" + c.item.name + "%N")
						end
					end
					file.put_string ("%N")
					file.put_string ("feature {" + data_class_name + "}%N%N")

					file.put_string ("	initialize%N")
					file.put_string ("		do%N")
					compound_statement.accept (eiffel_creator)
					file.put_string (eiffel_creator.buffer)
					file.put_string ("		end%N%N")

					file.put_string ("end%N")
					file.flush
					file.close
				else
					create slices_dir.make (transformer.project_settings.output_dir.name + "/init")
					if not slices_dir.exists then
						slices_dir.create_dir
					end

					create file.make_open_write (transformer.project_settings.output_dir.name + "/" + eiffel_file_name (initializer_class_name))
					file.put_string ("frozen class%N")
					file.put_string ("	" + initializer_class_name + "%N%N")
					file.put_string ("feature {" + data_class_name + "}%N%N")

					file.put_string ("	initialize%N")
					file.put_string ("		do%N")
					from
						j := 1
					until
						j > i
					loop
						file.put_string ("			(create {" + initializer_class_name + "_SLICE" + j.out + "}).initialize;%N")
						j := j + 1
					end
					file.put_string ("		end%N%N")
					file.put_string ("end%N")
					file.flush
					file.close

					from
						tmp := compound_statement.statements.twin
						j := 1
					until
						j > i
					loop
						create file.make_open_write (slices_dir.name + "/" + eiffel_file_name (initializer_class_name + "_slice" + j.out))
						file.put_string ("frozen class%N")
						file.put_string ("	" + initializer_class_name + "_SLICE" + j.out + "%N%N")

						file.put_string ("inherit%N")
						file.put_string ("	" + data_class_name + "%N")
						across classes as c loop
							-- FIXME: only actually required classes should be added
							if
								attached {OO_REGULAR_CLASS} c.item and
								c.item.is_system_source
							then
								file.put_string ("	" + c.item.name + "%N")
							end
						end
						file.put_string ("%N")
						file.put_string ("feature {" + initializer_class_name + "}%N%N")

						file.put_string ("	initialize%N")
						file.put_string ("		do%N")

						compound_statement.statements.wipe_out
						tmp.start
						tmp.split (slice_size)
						compound_statement.statements.fill (tmp.sublist)

						eiffel_creator.buffer.wipe_out
						compound_statement.accept (eiffel_creator)
						file.put_string (eiffel_creator.buffer)

						file.put_string ("		end%N%N")
						file.put_string ("end%N")
						file.flush
						file.close

						j := j + 1
					end
				end
			end
		end

	create_initializer_assignments (compound_statement: AST_COMPOUND_STATEMENT; target: AST_EXPRESSION; node: AST_EXPRESSION)
		require
			compound_statement /= Void
			target /= Void
			node /= Void
		local
			l_cur_target: AST_EXPRESSION
			l_fields: LIST [AST_STRUCT_FIELD]
			l_cursor: CURSOR
		do
			if attached {AST_INITIALIZER} node as tmp then
				if attached {AST_ARRAY_TYPE} tmp.type_binding then
					across tmp.elements as c loop
						l_cur_target := create {AST_POSTFIX_EXPRESSION}.make_array (
								target.ast_twin,
								create {AST_DECIMAL_CONSTANT}.make (c.target_index - 1, create {AST_BASIC_TYPE}.make (Ast_bt_int))
							)
						create_initializer_assignments (compound_statement, l_cur_target, c.item)
					end
				elseif attached {AST_STRUCT_TYPE} tmp.type_binding as st then
					l_fields := st.declaration_binding.fields
					l_cursor := l_fields.cursor
					l_fields.start
					across tmp.elements as c loop
						from
						until
							l_fields.item.identifier /= Void
						loop
							-- skip unnamed bitfields; they don't get a value
							l_fields.forth
						end
						l_cur_target := create {AST_POSTFIX_EXPRESSION}.make_point (target.ast_twin, l_fields.item.identifier, l_fields.item.type, l_fields.item)
						create_initializer_assignments (compound_statement, l_cur_target, c.item)
						l_fields.forth
					end
					l_fields.go_to (l_cursor)
				else
					across tmp.elements as c loop
						create_initializer_assignments (compound_statement, target, c.item)
					end
				end
			elseif
				attached {AST_STRING_CONSTANT} node as sc and
				attached {AST_ARRAY_TYPE} target.type_binding
			then
				sc.is_eiffel_string := True
				compound_statement.statements.extend (create {AST_EXPRESSION_STATEMENT}.make (create {AST_EIFFEL_EXPRESSION}.make_mixed (
						<<
							"across ",
							sc.ast_twin,
							" as c loop ",
							create {AST_PARENTHESIZED_EXPRESSION}.make (target.ast_twin),
							"[c.target_index - 1] := c.item.code.to_integer_8 end"
						>>,
						create {AST_BASIC_TYPE}.make (Ast_bt_void)
					)))
				compound_statement.statements.last.parent:= compound_statement
			else
				compound_statement.statements.extend (create {AST_EXPRESSION_STATEMENT}.make (create {AST_ASSIGNMENT_EXPRESSION}.make (target, node.ast_twin)))
				compound_statement.statements.last.parent:= compound_statement
			end
		end

	write_ecf_file (launcher_class_name: STRING; base_class_name: STRING; is_project: BOOLEAN)
		require
			launcher_class_name /= Void
			base_class_name /= Void
		local
			file: PLAIN_TEXT_FILE
		do
			create file.make_open_write (transformer.project_settings.output_dir.name + "/" + base_class_name.as_lower + ".ecf")
			file.put_string ("<?xml version=%"1.0%" encoding=%"ISO-8859-1%"?>%N")
			file.put_string ("<system name=%"" + base_class_name.as_lower + "%" ")
			if not is_project then
				file.put_string ("library_target=%"" + base_class_name.as_lower + "%" ")
			end
			file.put_string ("uuid=%"" + (create {UUID_GENERATOR}).generate_uuid.out + "%">%N")
			file.put_string ("	<target name=%"" + base_class_name.as_lower + "%">%N")
			if is_project then
				file.put_string ("		<root class=%"" + launcher_class_name + "%" feature=%"make%"/>%N")
			else
				file.put_string ("		<root all_classes=%"true%"/>%N")
			end
			file.put_string ("		<option warning=%"true%" syntax=%"provisional%">%N")
			file.put_string ("			<assertions precondition=%"true%" postcondition=%"true%" check=%"true%" invariant=%"true%" loop=%"true%" supplier_precondition=%"true%"/>%N")
			file.put_string ("		</option>%N")
			file.put_string ("		<setting name=%"console_application%" value=%"true%"/>%N")
			file.put_string ("		<external_include location=%"&quot;$ECF_CONFIG_PATH/include&quot;%"/>%N")
			across transformer.project_settings.include_dirs as c loop
				file.put_string ("		<external_include location=%"&quot;" + c.item + "&quot;%"/>%N")
			end
			if {PLATFORM}.is_unix then
				file.put_string ("		<external_object location=%"&quot;$ECF_CONFIG_PATH/objects/libffi.a&quot;%"/>%N")
			elseif {PLATFORM}.is_windows then
				file.put_string ("		<external_object location=%"&quot;$ECF_CONFIG_PATH/objects/libffi.lib&quot;%"/>%N")
			end
			across transformer.project_settings.libraries as c loop
				file.put_string ("		<external_library location=%"&quot;" + c.item + "&quot;%"/>%N")
			end
			if is_project then
				file.put_string ("		<precompile name=%"base_pre%" location=%"$ISE_PRECOMP\base.ecf%"/>%N")
			end
			file.put_string ("		<library name=%"base%" location=%"$ISE_LIBRARY\library\base\base.ecf%"/>%N")
			file.put_string ("		<cluster name=%"" + base_class_name.as_lower + "%" location=%".\%" recursive=%"true%">%N")
			file.put_string ("			<file_rule>%N")
			file.put_string ("				<exclude>/EIFGENs$</exclude>%N")
			file.put_string ("				<exclude>/CVS$</exclude>%N")
			file.put_string ("				<exclude>/.svn$</exclude>%N")
			file.put_string ("			</file_rule>%N")
			file.put_string ("		</cluster>%N")
			file.put_string ("	</target>%N")
			file.put_string ("</system>%N")
			file.flush
			file.close
		end

	write_project_launcher_class (main_function: AST_FUNCTION_DECLARATION; launcher_class_name, data_class_name: STRING)
		require
			main_function /= Void and then main_function.identifier.is_equal ("main")
			launcher_class_name /= Void
			data_class_name /= Void
		local
			l_has_return, l_has_argc, l_has_argc_argv: BOOLEAN
			l_file: PLAIN_TEXT_FILE
			l_instance_name: STRING_8
		do
			l_has_return := not (attached {AST_BASIC_TYPE} main_function.return_type as bt and then bt.type = Ast_bt_void)
			inspect main_function.parameter_list.parameters.count
			when 0 then
				-- no args
			when 1 then
				l_has_argc := True
			when 2 then
				l_has_argc_argv := True
			end

			create l_file.make_open_write (transformer.project_settings.output_dir.name + "/" + eiffel_file_name (launcher_class_name))
			l_file.put_string ("class%N")
			l_file.put_string ("	" + launcher_class_name + "%N%N")

			l_file.put_string ("inherit%N")
			if l_has_argc or l_has_argc_argv then
				l_file.put_string ("	ARGUMENTS%N")
			end
			l_file.put_string ("	" + data_class_name + "%N%N")

			l_file.put_string ("create%N")
			l_file.put_string ("	make%N%N")

			l_file.put_string ("feature {NONE} -- Initialization%N%N")

			l_file.put_string ("	make%N")
			if l_has_return or l_has_argc_argv then
				l_file.put_string ("		local%N")
				if l_has_argc_argv then
					l_file.put_string ("			i: INTEGER_32%N")
					l_file.put_string ("			argv: CE_ARRAY [CE_POINTER [INTEGER_8]]%N")
				end
				if l_has_return then
					l_file.put_string ("			return_value: INTEGER_32%N")
				end
			end
			l_file.put_string ("		do%N")
			if l_has_argc_argv then
				l_file.put_string ("			from%N")
				l_file.put_string ("				create argv.make (<<argument_count + 2>>)%N")
				l_file.put_string ("				i := 0%N")
				l_file.put_string ("			until%N")
				l_file.put_string ("				i > argument_count%N")
				l_file.put_string ("			loop%N")
				l_file.put_string ("				argv[i] := ce_string (argument_array[i])%N")
				l_file.put_string ("				i := i + 1%N")
				l_file.put_string ("			end%N")
			end

			l_instance_name := main_function.oo_feature.parent.name.as_lower + "_inst"
			if l_instance_name.starts_with ("p_") then
				l_instance_name := l_instance_name.substring (3, l_instance_name.count)
			end
			l_file.put_string ("			")
			if l_has_return then
				l_file.put_string ("return_value := ")
			end
			l_file.put_string (l_instance_name + ".main")
			if l_has_argc or l_has_argc_argv then
				l_file.put_string (" (argument_count + 1")
				if l_has_argc_argv then
					l_file.put_string (", argv")
				end
				l_file.put_string (")")
			end
			l_file.put_string ("%N")

			if l_has_return then
				l_file.put_string ("			if return_value /= 0 then%N")
				l_file.put_string ("				(create {EXCEPTIONS}).die (return_value)%N")
				l_file.put_string ("			end%N")
			end

			l_file.put_string ("		end%N%N")

			l_file.put_string ("end%N")
			l_file.flush
			l_file.close
		end

	write_project_data_class (
				data_class_name: STRING; classes: LIST [OO_CLASS];
				has_assembler: BOOLEAN; assembler_class_name: STRING;
				has_initializers: BOOLEAN; initializer_class_name: STRING
			)
		require
			data_class_name /= Void
			classes /= Void and then not classes.is_empty
		local
			l_file: PLAIN_TEXT_FILE
			l_instance_name: STRING_8
		do
			create l_file.make_open_write (transformer.project_settings.output_dir.name + "/" + eiffel_file_name (data_class_name))
			l_file.put_string ("deferred class%N")
			l_file.put_string ("	" + data_class_name + "%N%N")

			l_file.put_string ("inherit%N")
			l_file.put_string ("	CE_UTILITIES%N")
			if has_assembler then
				l_file.put_string ("	" + assembler_class_name + "%N")
			end
			l_file.put_string ("%N")

			l_file.put_string ("feature {NONE}%N%N")
			across classes as c loop
				if
					attached {OO_REGULAR_CLASS} c.item and
					not c.item.is_system_source
				then
					l_instance_name := c.item.name.as_lower + "_inst"
					if l_instance_name.starts_with ("p_") then
						l_instance_name := l_instance_name.substring (3, l_instance_name.count)
					end
					l_file.put_string ("	" + l_instance_name + ": " + c.item.name + "%N")
					l_file.put_string ("		once (%"PROCESS%")%N")
					l_file.put_string ("			create Result %N")
					l_file.put_string ("			" + data_class_name.as_lower + "_initialize%N")
					l_file.put_string ("		end%N%N")
				end
			end

			l_file.put_string ("feature {NONE}%N%N")
			l_file.put_string ("	" + data_class_name.as_lower + "_initialize%N")
			l_file.put_string ("		once (%"PROCESS%")")
			across classes as c loop
				if
					attached {OO_REGULAR_CLASS} c.item and
					not c.item.is_system_source
				then
					l_instance_name := c.item.name.as_lower + "_inst"
					if l_instance_name.starts_with ("p_") then
						l_instance_name := l_instance_name.substring (3, l_instance_name.count)
					end
					l_file.put_string ("%N")
					l_file.put_string ("			" + l_instance_name + ".do_nothing")
				end
			end
			if has_initializers then
				l_file.put_string (";%N")
				l_file.put_string ("			(create {" + initializer_class_name + "}).initialize")
			end
			l_file.put_string ("%N")
			l_file.put_string ("		end%N%N")

			l_file.put_string ("end%N")
			l_file.flush
			l_file.close
		end

	write_header_file (base_class_name: STRING)
		require
			base_class_name /= Void
		local
			d: DIRECTORY
			f: PLAIN_TEXT_FILE
			set: HASH_SET [STRING]
			printed: BOOLEAN
		do
			create d.make (transformer.project_settings.output_dir.name + "/include")
			if not d.exists then
				d.create_dir
			end
			create f.make_open_write (d.name + "/" + c_header_file (transformer.tree))
			f.put_string ("#ifndef " + base_class_name + "_H%N")
			f.put_string ("#define " + base_class_name + "_H%N%N")

			-- header files
			create set.make (16)
			across transformer.tree.declarations as c loop
				if
					c.item.source_info.is_system_source and
					not c.item.source_info.source_file.is_equal ("gcc_builtin.h")
				then
					set.extend (c.item.source_info.source_file)
				end
			end
			f.put_string ("#include <ffi.h>%N")
			across set.to_array as c loop
				f.put_string ("#include <" + c.item + ">%N")
			end
			f.put_string ("%N")

			-- Some header files might have macros that screw with our C code (happens in "links").
			-- "undef"ing stuff here:
			printed := False
			create set.make (1000)
			across transformer.tree.declarations as c loop
				if attached {AST_STRUCT_DECLARATION} c.item as tmp then
					across tmp.fields as c2 loop
						if attached c2.item.identifier as identifier then
							set.extend (identifier)
						end
					end
				elseif attached {AST_FUNCTION_DECLARATION} c.item as tmp and then tmp.is_extern then
					set.extend (tmp.identifier)
					across tmp.parameter_list.parameters as c2 loop
						if attached {AST_VARIABLE_DECLARATION} c2.item as tmp2 then
							set.extend (tmp2.orig_c_identifier)
						end
					end
				end
			end
			across set.to_array as c loop
				f.put_string ("#undef ")
				f.put_string (c.item)
				f.put_string ("%N")
				printed := True
			end

			-- enums
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if attached {AST_ENUM_DECLARATION} c.item as tmp and then tmp.source_info.is_project_source then
					f.put_string (tmp.c_code)
					f.put_string ("%N")
					printed := True
				end
			end

			-- struct forward declarations
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if
					attached {AST_STRUCT_DECLARATION} c.item as tmp and then
					(tmp.source_info.is_project_source or tmp.orig_c_identifier.starts_with ("__anon"))
				then
					if tmp.is_union then
						f.put_string ("union ")
					else
						f.put_string ("struct ")
					end
					f.put_string (tmp.orig_c_identifier)
					f.put_string (";%N")
					printed := True
				end
			end

			-- typedefs
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if
					attached {AST_STRUCT_DECLARATION} c.item as tmp and then
					tmp.source_info.is_project_source and then
					tmp.orig_c_identifier.starts_with ("__anon") and then
					tmp.typedef_name /= Void
				then
					f.put_string ("typedef ")
					if tmp.is_union then
						f.put_string ("union ")
					else
						f.put_string ("struct ")
					end
					f.put_string (tmp.orig_c_identifier + " " + tmp.typedef_name + ";%N")
					printed := True
				end
			end

			-- structs
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if
					attached {AST_STRUCT_DECLARATION} c.item as tmp and then
					(tmp.source_info.is_project_source or tmp.orig_c_identifier.starts_with ("__anon"))
				then
					f.put_string (tmp.c_code)
					f.put_string ("%N")
					printed := True
				end
			end

			-- external features from project headers
			if printed then f.put_string ("%N") end
			printed := False
			across transformer.tree.declarations as c loop
				if
					attached {AST_EIFFEL_FEATURE} c.item as tmp and then
					tmp.is_extern and then
					(
						tmp.source_info.is_project_source or
						tmp.orig_c_identifier.is_equal ("__builtin_va_copy")
					)
				then
					f.put_string (tmp.c_code)
					if attached {AST_VARIABLE_DECLARATION} tmp then
						f.put_string (";")
					end
					f.put_string ("%N")
					printed := True
				end
			end

			if printed then f.put_string ("%N") end
			f.put_string ("#endif%N")
			f.flush
			f.close
		end

	resources_directory: DIRECTORY
		local
			path: STRING
		do
			create Result.make ("../resources") -- we're in EiffelStudio
			if not Result.exists then
				create Result.make ("resources") -- it's a release
				if not Result.exists then
					path := Command_name -- e.g. "/foo/bar/C2Eif", "bar/C2Eif", "C2Eif"
					path := path.substring (1, path.last_index_of ('/', path.count)) -- "/foo/bar/", "bar/", ""
					create Result.make (path + "resources")
					if not Result.exists then
						Result := Void
					end
				end
			end
		end

	copy_utilities (src_path, dest_path: STRING)
		require
			src_path /= Void
			dest_path /= Void
		local
			src_directory, dest_directory: DIRECTORY
			src_file, dest_file: RAW_FILE
		do
			create src_directory.make (src_path)
			if src_directory.exists then
				create dest_directory.make (transformer.project_settings.output_dir.name + "/" + dest_path)
				if not dest_directory.exists then
					dest_directory.create_dir
				end

				across src_directory.linear_representation as c loop
					create src_file.make (src_directory.name + "/" + c.item)
					if not src_file.is_directory then -- ignore `.' and `..'
						src_file.open_read
						create dest_file.make_open_write (dest_directory.name + "/" + c.item)
						src_file.copy_to (dest_file)
						dest_file.flush
						dest_file.close
						src_file.close
					end
				end
			else
				log ("Can't copy utilities from %"" + src_path + "%": directory doesn't exist")
			end
		end

	write_ce_settings_class (dest_path: DIRECTORY)
		require
			dest_path /= Void and then dest_path.exists
		local
			f: PLAIN_TEXT_FILE
		do
			create f.make_open_write (dest_path.name + "/ce_settings.e")
			f.put_string ("class%N")
			f.put_string ("	CE_SETTINGS%N")
			f.put_string ("%N")
			f.put_string ("feature {ANY}%N")
			f.put_string ("%N")
			f.put_string ("	Disable_bounds_checking: BOOLEAN = " + transformer.project_settings.disable_bounds_checking.out + "%N")
			f.put_string ("%N")
			f.put_string ("end%N")
			f.flush
			f.close
		end

	remove_const_qualifiers (a_node: AST_NODE): AST_NODE
		local
			l_visitor: AST_AGENT_VISITOR
		do
			Result := a_node.ast_twin

			create l_visitor
			l_visitor.post_visit_agent := agent (ag_node: AST_NODE)
				do
					if attached {AST_TYPE} ag_node as tmp and then tmp.is_const then
						tmp.is_const := False
					end
				end
			Result.accept (l_visitor)
		end

end
