class
	NAME_UTILITIES

inherit
	ANTLR_AST_TYPE_CONSTANTS

feature {ANY}

	eiffel_file_name (c_name: STRING): STRING
			-- `c_name' can be
			-- o a file name (e.g. "micro_httpd.cil.adapted.c") or s
			-- o a struct/union name (e.g. "_IO_FILE")
			-- returns e.g. "io_file.e" for "_IO_FILE"
		require
			c_name /= Void and then not c_name.is_empty
		do
			Result := eiffel_class_name (c_name).as_lower + ".e"
		end

	eiffel_class_name (c_name: STRING): STRING
			-- `c_name' can be
			-- o a file name (e.g. "micro_httpd.cil.adapted.c") or
			-- o a struct/union name (e.g. "_IO_FILE")
			-- returns e.g. "MICRO_HTTPD" for "micro_httpd.cil.adapted.c"
		require
			c_name /= Void and then not c_name.is_empty
		do
			create Result.make_from_string (c_name)
			from until not Result.starts_with ("_") loop
				Result := Result.substring (2, Result.count)
			end
			from until not Result.starts_with ("../") loop
				Result := Result.substring (4, Result.count)
			end
			if Result.has ('.') then
				Result := Result.substring (1, Result.index_of ('.', 1) - 1)
			end
			Result.replace_substring_all ("-", "_")
			Result.replace_substring_all ("/", "_")
			Result.to_upper

			if Result[1].is_digit then
				Result := "C2EIF_" + Result
			end
		end

	c_header_file (node: AST_TRANSLATION_UNIT): STRING
		require
			node /= Void
		do
			Result := eiffel_class_name (node.source_file_name).as_lower + ".h"
		end

feature {NONE}

	multiply (s: STRING; count: INTEGER): STRING
		require
			count >= 0
		do
			if count = 0 then
				Result := ""
			else
				Result := s
				Result.multiply (count)
			end
		ensure
			Result /= Void
		end

end
